# An R function to calculate the X,Y,Z position of an individual using stereographic
# image analysis. See http://exon.niaid.nih.gov/3dswarms/ for more details.

# N.C. Manoukis
# NIAID/NIH
# manoukisn@niaid.nih.gov

# Note: The X,Y, Z positions are calculated based on the "standard model" for
# stereophotography: 2 parallel cameras with the same characteristics (f, sensor 
# size).

# required inputs:
# xl= vector with x position (in pixels) in the LEFT image of each object to be localized
# xr= vector with x position (in pixels) in the RIGHT image of each object to be localized
# yl= vector with y position (in pixels) in the LEFT image of each object to be localized
# yr= vector with y position (in pixels) in the RIGHT image of each object to be localized
# z_method= calculate Z by triangulation (0;default) or by regression based on calibration data (1)
# calib_set= (only needed if z_method = 1) calibration data.frame. Needs to have variables Xdiff and Dist
# cam= which camera to use for X,Y calculation ('L' or 'R').
# poly_corr= use distortion correction based on polynomial (deg=2) regression performed on 
#	     calibration images; default=F. To use this option, set the values for the polynomial
#            within the script, below.

# changelog
# ver 0.1: Sep 2007- first version
# ver 0.2: Mar 2008- include polynomial correction
# ver 0.3: Nov 2008- beefed up documentation

calcxyz <- function (xl,xr,yl,yr,z_method=0,calib_set,cam,poly_corr=F){

	###############################################
	## Set values for your camera geometry below ##
	###############################################

	b<-19/100 # baseline, m
	
	fh<-5.4/1000 # focal length in m
        fv<-(4/3)*5.4/1000 # ditto for vertical, may vary with proper calibration
	
	# sensor description below for my cam = 1/2.9 inches; this will almost certainly not work for you

        ch<-720 # optical center, horizontal
        cv<-540 # optical center, vertical

        rh <- 1440 # H resolution (px)
        rv <- 1080 # V resolution (px)

        sh <- 4.8/1000 # sensor h size, in m
        sv <- 3.6/1000 # sensor v size, in m
	
	#sh <- 5.27/1000 # sensor h size, in m
	#sv <- 3.96/1000 # sensor v size, in m
	
	## Set polynomial correction values below
	if (poly_corr==T){
		# use your own estimated correction values here
		intercept <- 0.054454;
		f1<- 1.227194; # estimate for linear term
		f2<- -0.064525; # estimate for sqared term
	}
	
	###############################################
	## Set values for your camera geometry above ##
	###############################################
	
	# very simple error checks
	lxl<-length(xl);lxr<-length(xr);lyl<-length(yl);lyr<-length(yr)
	if(lxl!=lxr|lxl!=lyl|lxl!=lyr|lxr!=lyl|lxr!=lyr|lyl!=lyr){
		warning('length of all input vectors is not equal')
		warning('check inputs')
	}
	
	if(cam!='L'){
		if(cam!='R'){
			warning('I don\'t understand master camera specification- should be L or R')
		}
	}
	
	#initialize counter, set up output vectors
	count <- 1; len <- lxl
	X<- c(); Y<- c(); Z<-c()
	
	# calculate Z; needed to calculate X,Y
	if (z_method==0){ # triangulation
		if (poly_corr==T){
			Z<-calc_Z(xl,xr,b,fh,ch,rh,sh,poly_corr=T,intercept,f1,f2)
			}
		else{
			Z<-calc_Z(xl,xr,b,fh,ch,rh,sh)
			}
		
		}
	else {
		if (z_method==1){ #linear regression with training data
			cmod<-lm(1/calib_set$Dist*100~calib_set$Xdiff)
			n<-data.frame(Xdiff=xl-xr)
			Z<-(predict.lm(cmod,n)/100)^-1
		}
		else {
		warning('I don\'t know which Z method to choose')
		}
	}
	# X,Y loop
	while (count <= len){# using similar triangles, calculate actual X,Y
		if (cam=='L'){
			f<-striang(xl[count],yl[count],Z[count],fh,fv,ch,cv,rh,rv,sv,sh)
			}
		if (cam=='R'){
			f<-striang(xr[count],yr[count],Z[count],fh,fv,ch,cv,rh,rv,sv,sh)
			}
		X<-c(X,f$x)
		Y<-c(Y,f$y)
		count <- count+1
	}
	final_out <- data.frame(cbind(X=X,Y=Y,Z=Z))
	
	
	return (final_out)

}


## Calculates the distance of a stereo point from the cameras. This is *not*
## along the ray that goes from the projection center, through the point on 
## the image plane and out to the actual object; rather, it is the distance 
## along the Z axis (depth)
calc_Z<-function(ul,ur,b,f,ch,rh,sh,poly_corr=F,int,f1,f2){
	num<-b*f
	uml <- ((ul-ch)/rh)*sh
	umr <- ((ur-ch)/rh)*sh
	Z<-num/(uml-umr);
	if (poly_corr==T){
		Z<-int+(f1*Z)+(f2*Z^2);
		}
	return (Z)
}

## Calculate X,Y in real-world units. i.e., Z gives
## X, Y displacement from the optic center
striang<-function(u,v,z,fh,fv,ch,cv,rh,rv,sv,sh){
        # get u,v in m; always off of center
        um <- ((u-ch)/rh)*sh
        vm <- ((v-cv)/rv)*sv

	# get X,Y
        Xs <- (z/fh)*um
        Ys <- -(z/fv)*vm # -ve to correct for pinhole proj

        # catch case of mosquito at optical center
        if (um==0){X<-um}
        if (vm==0){Y<-vm}

        return (data.frame(cbind(x=Xs,y=Ys)))
}

